/*

  libkohn_rtsp - Copyright 2011 by Michael Kohn
  http://www.mikekohn.net/
  mike@mikekohn.net

  Source distributed under the LGPL license.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <poll.h>
#include <time.h>
#include <pthread.h>

#include "kohn_rtsp.h"
#include "kohn_rtp_thread.h"

#define MAKE_WORD32(a) ((buffer[a]<<24)|(buffer[a+1]<<16)|(buffer[a+2]<<8)|(buffer[a+3]))
#define MAKE_BYTE4(a,b) buffer[b]=a>>24; buffer[b+1]=(a>>16)&0xff; buffer[b+2]=(a>>8)&0xff; buffer[b+3]=a&0xff;

static int running=0;
static int dont_init=0;
static struct _kohn_rtsp *kohn_rtsp_head=NULL;
static struct _kohn_rtsp *kohn_rtsp_add=NULL;
#ifdef WINDOWS
CRITICAL_SECTION thread_mutex;
#else
pthread_mutex_t thread_mutex;
pthread_mutexattr_t thread_mutex_attr;
#endif

static void kohn_rtsp_rtp_thread(int *params)
{
struct _kohn_rtsp *kohn_rtsp_prev;
struct _kohn_rtsp *kohn_rtsp_curr;
struct pollfd fds[KOHN_RTSP_MAX_CHANNELS*2];
struct sockaddr_in dest_addr;
unsigned char buffer[KOHN_RTSP_BUFFER_RTP_SIZE];
unsigned char *rtp;
struct _buffered_rtp *buffered_rtp=NULL;
unsigned int buffered_rtp_len;     // FIXME - useless? (could be used to grow)
unsigned int buffered_rtp_ptr=0;   // Points to next free rtp area
unsigned int buffered_rtp_count=0; // Number of used slots
int offset;
nfds_t nfds;
int ret;
time_t curr_time;

  running=1;

  memset(&dest_addr, 0, sizeof(dest_addr));
  dest_addr.sin_family=AF_INET;

  while(running)
  {
    curr_time=time(NULL);

    // If new channels need to be added, do it!
    if (kohn_rtsp_add!=NULL)
    {
#ifndef WINDOWS
      pthread_mutex_lock(&thread_mutex);
#else
      EnterCriticalSection(&thread_mutex);
#endif

      kohn_rtsp_curr=kohn_rtsp_add;
      while(kohn_rtsp_curr!=NULL)
      {
        if (kohn_rtsp_curr->mode>0)
        {
          if (buffered_rtp==NULL)
          {
#ifdef DEBUG
            printf("Allocating rtp buffer memory pool.\n");
#endif
            buffered_rtp=(struct _buffered_rtp *)malloc(sizeof(struct _buffered_rtp)*KOHN_RTSP_BUFFER_RTP_COUNT);
            memset(buffered_rtp,0,sizeof(struct _buffered_rtp)*KOHN_RTSP_BUFFER_RTP_COUNT);
            buffered_rtp_len=KOHN_RTSP_BUFFER_RTP_COUNT;
            buffered_rtp_ptr=0;
            buffered_rtp_count=0;

            int n;
            for (n=0; n<KOHN_RTSP_BUFFER_RTP_COUNT; n++)
            { buffered_rtp[n].kohn_rtsp=NULL; }
          }

          if (kohn_rtsp_curr->mode==KOHN_RTP_MODE_MARKER)
          {
            kohn_rtsp_curr->frame_data_len=KOHN_RTSP_FRAME_DATA_START_LEN;
            kohn_rtsp_curr->frame_data_ptr=0;
            kohn_rtsp_curr->frame_data=(unsigned char *)malloc(KOHN_RTSP_FRAME_DATA_START_LEN);
          }
        }

        kohn_rtsp_curr=kohn_rtsp_curr->next;
      }

      if (kohn_rtsp_head==NULL)
      {
        kohn_rtsp_head=kohn_rtsp_add;
      }
        else
      {
        kohn_rtsp_curr=kohn_rtsp_head;
        while(kohn_rtsp_curr->next!=NULL)
        { kohn_rtsp_curr=kohn_rtsp_curr->next; }
        kohn_rtsp_curr->next=kohn_rtsp_add;
      }
      kohn_rtsp_add=NULL;

#ifndef WINDOWS
      pthread_mutex_unlock(&thread_mutex);
#else
      LeaveCriticalSection(&thread_mutex);
#endif
    }

    // Poll for data
    kohn_rtsp_curr=kohn_rtsp_head;
    nfds=0;
    while(kohn_rtsp_curr!=NULL)
    {
      // FIXME - This really only needs to update if something changed.
      fds[nfds].fd=kohn_rtsp_curr->rtp_sockfd;
      fds[nfds].events=POLLIN;
      fds[nfds].revents=0;  // FIXME - check if this really needs to be cleared
      fds[nfds+1].fd=kohn_rtsp_curr->rtcp_sockfd;
      fds[nfds+1].events=POLLIN;
      fds[nfds+1].revents=0;// FIXME - check if this really needs to be cleared
      kohn_rtsp_curr=kohn_rtsp_curr->next;
      nfds+=2;
    }

    ret=poll(fds,nfds,KOHN_RTSP_POLL_TIMEOUT_MS);

    // Timeout occured.
    //if (ret==0)
    //{
    //  continue;
    //}

    // Check poll and process or remove channels if needed
    kohn_rtsp_prev=NULL;
    kohn_rtsp_curr=kohn_rtsp_head;
    nfds=0;
    while(kohn_rtsp_curr!=NULL)
    {
      if (fds[nfds].revents==POLLERR ||
          fds[nfds].revents==POLLHUP ||
          fds[nfds+1].revents==POLLERR ||
          fds[nfds+1].revents==POLLHUP)
      {
#ifdef DEBUG
        printf("RTP Thread: context %p died. (%x)\n", kohn_rtsp_curr->context,fds[nfds].revents);
#endif
        kohn_rtsp_curr->inuse=2;
      }
        else
      if (fds[nfds].revents==POLLIN)
      {
#ifdef SUPER_DEBUG
        printf("RTP Thread: Receive RTP data on id %d.\n",
               kohn_rtsp_curr->id);
#endif
        kohn_rtsp_curr->packet_count++;

        if ((kohn_rtsp_curr->mode&1)==0)
        {
          rtp=buffer;
        }
          else
        {
          rtp=buffered_rtp[buffered_rtp_ptr].rtp;
        }

        ret=recvfrom(kohn_rtsp_curr->rtp_sockfd, rtp, KOHN_RTSP_BUFFER_RTP_SIZE, 0, NULL, 0);

#ifdef DEBUG
printf("UDP packet size ret=%d %d  %d\n",ret,kohn_rtsp_curr->rtp_sockfd,buffered_rtp_ptr);
#endif

        if (ret>0)
        {
#ifdef DEBUG
          printf("RTP Packet --- %d bytes\n", ret);
          printf("     Version: %d\n", rtp[0]>>6);
          printf("     Padding: %d\n", (rtp[0]>>5)&1);
          printf("   Extension: %d\n", (rtp[0]>>4)&1);
          printf("  CSRC Count: %d\n", rtp[0]&0x0f);
          printf("      Marker: %d\n", rtp[1]>>7);
          printf("Payload Type: %d\n", rtp[1]&0x7f);
          printf("Sequence Num: %u\n", (rtp[2]<<8)|rtp[3]);
          printf("   Timestamp: %u\n", (rtp[4]<<24)|(rtp[5]<<16)|(rtp[6]<<8)|rtp[7]);
          printf("        SSRC: %d\n", (rtp[8]<<24)|(rtp[9]<<16)|(rtp[10]<<8)|rtp[11]);
#endif
          offset=12+((rtp[0]&0x0f)*4);
          if (((rtp[0]>>4)&1)!=0)
          {
            offset+=((rtp[offset+2]<<8)|rtp[offset+3])+4;
          }

          if (offset<ret)
          {
            // FIXME - move this upward so the offset doesn't have to be
            //         calculated for no reason
            if (kohn_rtsp_curr->mode==KOHN_RTP_MODE_RTP)
            {
              kohn_rtsp_curr->rtp_callback(rtp, ret, kohn_rtsp_curr->context);
            }
              else
            if (kohn_rtsp_curr->mode==KOHN_RTP_MODE_RAW)
            {
              kohn_rtsp_curr->rtp_callback(rtp+offset, ret-offset, kohn_rtsp_curr->context);
            }
              else
            {
              int seq=(rtp[2]<<8)|rtp[3];
              int marker=rtp[1]>>7;

#ifdef DEBUG
              if (kohn_rtsp_curr->seq!=seq)
              {
                printf("Out of order (%p) %d and %d\n", kohn_rtsp_curr->context, kohn_rtsp_curr->seq, seq);
              }
#endif

              if (seq<kohn_rtsp_curr->seq)
              {
#ifdef DEBUG
                printf("Oh crap :(  seq %d is less than current %d and marker=%d\n", seq, kohn_rtsp_curr->seq, marker);
#endif
/*

                if (kohn_rtsp_curr->mode==KOHN_RTP_MODE_MARKER && marker==1)
                {
                  // This is all fucked up.  Might as well dump the packet.
                  kohn_rtsp_curr->callback(kohn_rtsp_curr->id, kohn_rtsp_curr->frame_data, kohn_rtsp_curr->frame_data_ptr);
                  kohn_rtsp_curr->frame_data_ptr=0;
                }
*/

                kohn_rtsp_curr=kohn_rtsp_curr->next;
                nfds+=2;
                continue;
              }

              buffered_rtp[buffered_rtp_ptr].next=NULL;
              buffered_rtp[buffered_rtp_ptr].kohn_rtsp=kohn_rtsp_curr;
              buffered_rtp[buffered_rtp_ptr].seq=seq;
              buffered_rtp[buffered_rtp_ptr].len=ret-offset;
              buffered_rtp[buffered_rtp_ptr].marker=marker;
              buffered_rtp[buffered_rtp_ptr].offset=offset;
              struct _buffered_rtp *buffered_rtp_curr;

              // Que RTP packet up in channels RTP buffer linked list
              if (kohn_rtsp_curr->rtp_head==NULL)
              {
                kohn_rtsp_curr->rtp_head=&buffered_rtp[buffered_rtp_ptr];
              }
                else
              {
                buffered_rtp_curr=kohn_rtsp_curr->rtp_head;
                while(1)
                {
                  if (buffered_rtp_curr->next==NULL)
                  {
                    buffered_rtp_curr->next=&buffered_rtp[buffered_rtp_ptr];
                    break;
                  }

                  if (buffered_rtp_curr->seq<seq &&
                      seq<buffered_rtp_curr->next->seq)
                  {
                    buffered_rtp[buffered_rtp_ptr].next=buffered_rtp_curr->next;
                    buffered_rtp_curr->next=&buffered_rtp[buffered_rtp_ptr];
                    break;
                  }
                  buffered_rtp_curr=buffered_rtp_curr->next;
                }
              }

              kohn_rtsp_curr->buffered_count++;

              buffered_rtp_count++;

              // Attempt to unque it or anything we can
              buffered_rtp_curr=kohn_rtsp_curr->rtp_head;

              if (kohn_rtsp_curr->mode==KOHN_RTP_MODE_ORDER)
              {
                // Sent out any rtp packets that are in order
                while(buffered_rtp_curr!=NULL)
                {
                  if (buffered_rtp_curr->seq<=kohn_rtsp_curr->seq ||
                      kohn_rtsp_curr->buffered_count>=KOHN_RTSP_OUT_OF_ORDER_MAX ||
                      buffered_rtp_count>=KOHN_RTSP_BUFFER_RTP_COUNT)
                  {
                    kohn_rtsp_curr->seq=buffered_rtp_curr->seq+1;
                    kohn_rtsp_curr->rtp_callback(buffered_rtp_curr->rtp+buffered_rtp_curr->offset, buffered_rtp_curr->len, kohn_rtsp_curr->context);
                    buffered_rtp_curr->kohn_rtsp=NULL;
                    buffered_rtp_count--;
                    buffered_rtp_curr=buffered_rtp_curr->next;
                  }
                    else
                  { break; }
                }
              }
                else
              if (kohn_rtsp_curr->mode==KOHN_RTP_MODE_MARKER)
              {
                // Buffer up any RTP packets in order until markers
                while(buffered_rtp_curr!=NULL)
                {
                  if (buffered_rtp_curr->seq<=kohn_rtsp_curr->seq ||
                      kohn_rtsp_curr->buffered_count>=KOHN_RTSP_OUT_OF_ORDER_MAX ||
                      buffered_rtp_count>=KOHN_RTSP_BUFFER_RTP_COUNT)
                  {
                    kohn_rtsp_curr->seq=buffered_rtp_curr->seq+1;

                    // Grow buffer maybe
                    if (kohn_rtsp_curr->frame_data_ptr+buffered_rtp_curr->len>
                        kohn_rtsp_curr->frame_data_len)
                    {
#ifdef DEBUG
                      printf("context %p> frame_data_len too small %d.\n", kohn_rtsp_curr->context, kohn_rtsp_curr->frame_data_len);
#endif
                      if (kohn_rtsp_curr->frame_data_len<KOHN_RTSP_FRAME_DATA_MAX_LEN)
                      {
                        kohn_rtsp_curr->frame_data_len+=KOHN_RTSP_FRAME_DATA_GROW_LEN;
                        kohn_rtsp_curr->frame_data=(unsigned char *)realloc(kohn_rtsp_curr->frame_data, kohn_rtsp_curr->frame_data_len);
                      }
                        else
                      {
#ifdef DEBUG
                      printf("context %p> frame_data_len can't grow.  Calling back with bad data.\n", kohn_rtsp_curr->context);
#endif
                        // If this happens it's bad
                        kohn_rtsp_curr->rtp_callback(kohn_rtsp_curr->frame_data, kohn_rtsp_curr->frame_data_ptr, kohn_rtsp_curr->context);
                        kohn_rtsp_curr->frame_data_ptr=0;
                      }
                    }


#ifdef DEBUG
                    // FIXME - Get rid of this debug later
                    if (kohn_rtsp_curr->fourcc[0]=='h' ||
                        kohn_rtsp_curr->fourcc[1]=='2' ||
                        kohn_rtsp_curr->fourcc[2]=='6' ||
                        kohn_rtsp_curr->fourcc[3]=='4')
                    {
                    unsigned char *a=buffered_rtp_curr->rtp+buffered_rtp_curr->offset;
                    int type=a[0]&0x1f;
                    printf("NAL Unit Type=%02x  F=%d NRI=%d Type=%d\n", a[0], a[0]>>7, (a[0]>>5)&0x03, a[0]&0x1f);
                    if (type==28)
                    {
                      printf("FU-A header=%02x S=%d E=%d R=%d Type=%d\n", a[1],
a[1]>>7, (a[1]>>6)&0x01, (a[1]>>5)&0x01, a[1]&0x1f);
                      buffered_rtp_curr->offset+=2;
                      buffered_rtp_curr->len-=2;
                    }
                    }
#endif

                    memcpy(kohn_rtsp_curr->frame_data+kohn_rtsp_curr->frame_data_ptr, buffered_rtp_curr->rtp+buffered_rtp_curr->offset, buffered_rtp_curr->len);
                    kohn_rtsp_curr->frame_data_ptr+=buffered_rtp_curr->len;

                    if (buffered_rtp_curr->marker==1)
                    {
                      kohn_rtsp_curr->rtp_callback(kohn_rtsp_curr->frame_data, kohn_rtsp_curr->frame_data_ptr, kohn_rtsp_curr->context);
                      kohn_rtsp_curr->frame_data_ptr=0;
                    }

                    buffered_rtp_curr->kohn_rtsp=NULL;
                    buffered_rtp_count--;
                    buffered_rtp_curr=buffered_rtp_curr->next;
                  }
                    else
                  { break; }
                }
              }

              kohn_rtsp_curr->rtp_head=buffered_rtp_curr;

              // Find next free RTP buffer block
              int n;
              for (n=0; n<buffered_rtp_len; n++)
              {
                if (buffered_rtp[buffered_rtp_ptr].kohn_rtsp==NULL) break;
                buffered_rtp_ptr++;
                if (buffered_rtp_ptr>=KOHN_RTSP_BUFFER_RTP_COUNT)
                { buffered_rtp_ptr=0; }
              }

              if (n==buffered_rtp_len)
              {
                printf("Buffer is full.  Dying at %s:%d.\n", __FILE__, __LINE__);
                exit(1);
              }
            }
          }
        }
      }

      if (fds[nfds+1].revents==POLLIN)
      {
        kohn_rtsp_curr->rtcp_sr_time=curr_time;

        ret=recvfrom(kohn_rtsp_curr->rtcp_sockfd, buffer, KOHN_RTSP_BUFFER_RTP_SIZE, 0, NULL, 0);
#ifdef SUPER_DEBUG
        printf("RTP Thread: Receive %d RTCP databytes on id %d.\n", ret, kohn_rtsp_curr->id);
#endif

#ifdef DEBUG
        // http://www.networksorcery.com/enp/protocol/rtcp.htm
        printf("RTCP Packet --- %d bytes\n", ret);
        printf("     Version: %d\n", buffer[0]>>6);
        printf("     Padding: %d\n", (buffer[0]>>5)&1);
        printf("Report Count: %d\n", buffer[0]&0x1f);
        printf("        Type: %d\n", buffer[1]);
        printf("      Length: %d\n", (buffer[2]<<8)+buffer[3]);
        printf("        SSRC: %u\n", MAKE_WORD32(4));
        printf("NTP time MSW: %u\n", MAKE_WORD32(8));
        printf("NTP time LSW: %u\n", MAKE_WORD32(12));
        printf("    RTP time: %u\n", MAKE_WORD32(16));
        printf("snd pckt cnt: %d\n", MAKE_WORD32(20));
        printf(" snd oct cnt: %d\n", MAKE_WORD32(24));
        if (ret>24)
        {
          printf("----\n");
          printf("      SSRC_1: %u\n", MAKE_WORD32(28));
          printf("  fract lost: %u\n", buffer[32]);
          printf(" packet lost: %u\n", (buffer[33]<<16)|(buffer[34]<<8)|buffer[35]);
          printf(" extd hi seq: %u\n", MAKE_WORD32(36));
          printf("      Jitter: %u\n", MAKE_WORD32(40));
          printf("Last SR(LSR): %u\n", MAKE_WORD32(44));
          printf("Dlay SR(DSR): %u\n", MAKE_WORD32(48));
        }
        if (ret>48)
        {
          printf("----\n");
          printf("      SSRC_2: %u\n", MAKE_WORD32(52));
          printf("            : %u\n", MAKE_WORD32(56));
        }
#endif
        kohn_rtsp_curr->ssrc_1=MAKE_WORD32(28);
        kohn_rtsp_curr->last_sr=MAKE_WORD32(44);

        if (kohn_rtsp_curr->rtcp_callback!=NULL)
        {
          kohn_rtsp_curr->rtp_callback(buffer, ret, kohn_rtsp_curr->context);
        }
      }

      // Drop connection if no RTCP SR packet in too long of time
      if (curr_time-kohn_rtsp_curr->rtcp_sr_time>KOHN_RTSP_RTCP_SR_INTERVAL_S)
      {
        kohn_rtsp_curr->inuse=4;
      }

      // Send out RTCP RR packet to let camera know we are alive
      if (curr_time-kohn_rtsp_curr->rtcp_rr_time>KOHN_RTSP_RTCP_RR_INTERVAL_S)
      {
        #define RTCP_PACKET_LEN 7
        // FIXME - This isn't correct info
        buffer[0]=(0x02<<6);
        buffer[1]=201;                              // Type: RR
        buffer[2]=0;                                // Len MSB
        buffer[3]=RTCP_PACKET_LEN;                  // Len LSB
        MAKE_BYTE4(kohn_rtsp_curr->ssrc, 4);
        MAKE_BYTE4(kohn_rtsp_curr->ssrc_1, 8);      // SSRC_1 ?
        buffer[12]=kohn_rtsp_curr->fraction_lost; 
        buffer[13]=0;                               // packets lost
        buffer[14]=0; 
        buffer[15]=0; 
        MAKE_BYTE4(kohn_rtsp_curr->seq, 16);
        MAKE_BYTE4(0, 20);                          // Jitter
        MAKE_BYTE4(kohn_rtsp_curr->last_sr, 24);    // Last SR
        MAKE_BYTE4(0, 28);                          // Delay SR

        dest_addr.sin_port=kohn_rtsp_curr->server_ports+1;
        memcpy(&dest_addr.sin_addr, &kohn_rtsp_curr->sin_addr, sizeof(struct in_addr));
        //if (inet_aton("192.168.1.100", &dest_addr.sin_addr)==0)
        //{
        //  printf("Bad destination address\n");
        //  exit(1);
        //}

        sendto(kohn_rtsp_curr->rtcp_sockfd, buffer, (RTCP_PACKET_LEN+1)*4, 0, (struct sockaddr *)&dest_addr, sizeof(dest_addr));
      }

      if (kohn_rtsp_curr->inuse!=1)
      {
#ifdef DEBUG
        printf("Removing %p from RTP thread.\n", kohn_rtsp_curr->context);
#endif
        // Remove any buffered packets for this channel.
        struct _buffered_rtp *buffered_rtp_curr=kohn_rtsp_curr->rtp_head;
        while(buffered_rtp_curr!=NULL)
        {
          buffered_rtp_curr->kohn_rtsp=NULL;
          buffered_rtp_curr=buffered_rtp_curr->next;
          buffered_rtp_count--;
        }

        close(kohn_rtsp_curr->rtp_sockfd);
        close(kohn_rtsp_curr->rtcp_sockfd);
        free(kohn_rtsp_curr->frame_data);
        kohn_rtsp_curr->frame_data=0;
        kohn_rtsp_curr->inuse=0;

        if (kohn_rtsp_prev==NULL)
        {
          kohn_rtsp_head=kohn_rtsp_curr->next;
        }
          else
        {
          kohn_rtsp_prev->next=kohn_rtsp_curr->next;
        }
      }
        else
      {
        kohn_rtsp_prev=kohn_rtsp_curr;
      }

      kohn_rtsp_curr=kohn_rtsp_curr->next;
      nfds+=2;
    }

    if (kohn_rtsp_head==NULL) { running=0; }
  }

#ifdef DEBUG
  printf("RTP thread is dying.\n");
#endif

  if (buffered_rtp!=NULL) free(buffered_rtp);
  dont_init=0;
}

int kohn_rtsp_is_rtp_thread_running()
{
  return running;
}

int kohn_rtsp_start_rtp(struct _kohn_rtsp *kohn_rtsp)
{
  kohn_rtsp->next=NULL;
  kohn_rtsp->rtcp_sr_time=time(NULL);
  kohn_rtsp->rtcp_rr_time=time(NULL);

#ifdef DEBUG
  printf("kohn_rtsp_start_rtp: %p\n", kohn_rtsp);
#endif

  if (running==0)
  {
    //kohn_rtsp_head=kohn_rtsp;
    kohn_rtsp_add=kohn_rtsp;

    if (dont_init==0)
    {
#ifndef WINDOWS
      pthread_mutexattr_init(&thread_mutex_attr);
      pthread_mutex_init(&thread_mutex, &thread_mutex_attr);
#else
      InitializeCriticalSection(&thread_mutex);
#endif
      dont_init=1;
    }

#ifndef WINDOWS
    pthread_t pid;
    pthread_create(&pid,NULL,(void*)kohn_rtsp_rtp_thread,NULL);
#else
    _beginthread((void*)kohn_rtsp_rtp_thread,0,NULL);
#endif
  }
    else
  {
#ifndef WINDOWS
  pthread_mutex_lock(&thread_mutex);
#else
  EnterCriticalSection(&thread_mutex);
#endif

    if (kohn_rtsp_add==NULL)
    {
      kohn_rtsp_add=kohn_rtsp;
    }
      else
    {
      struct _kohn_rtsp *kohn_rtsp_add_curr=kohn_rtsp_add->next;
      while (kohn_rtsp_add_curr->next!=NULL)
      { kohn_rtsp_add_curr=kohn_rtsp_add_curr->next; }
      kohn_rtsp_add_curr->next=kohn_rtsp;
    }

#ifndef WINDOWS
    pthread_mutex_unlock(&thread_mutex);
#else
    LeaveCriticalSection(&thread_mutex);
#endif
  }

  return 0;
}


